# meds/event.py
#
#

""" object representing an event, uses a string as the default value and is able to be parsed for commands. """

from meds.utils.tijd import to_day, get_day, to_time
from meds.utils.trace import get_exception
from meds.errors import ENOTSET, ENODATE
from meds.utils.getters import days
from meds.utils.join import sj, j
from meds.object import Object
from meds.cfg import cfg

import logging
import types
import time
import os

class Event(Object):

    def __getattr__(self, name):
        try: return super().__getattr__(name)
        except AttributeError: self[name] = ""
        return self[name]                

    def announce(self, txt):
        """ announce on all available channels on all available bots. """
        from meds.core import fleet
        for bot in fleet:
            bot.announce(txt)

    def dispatch(self):
        """ execute all functione registered in self._funcs list. """
        for func in self._funcs:
            logging.info("! exec %s (%s)" % (str(func), self.origin))
            func(self)

    def display(self, obj, keys=[], txt=""):
        """ display the content of an object. """
        res = ""
        if not keys: keys = obj.keys()
        for key in sorted(keys):
            val = getattr(obj, key, None)
            if val: res += " " + str(val)
        try: res += " - %s" % days(obj)
        except ENODATE: pass
        txt += res
        if txt: self.reply(txt.strip())

    def join(self, *args, **kwargs):
        """ join threads started while handling this event. """
        if self._thr: return self._thr.join(*args, **kwargs)

    def ok(self, txt=""):
        """ reply with 'ok'. """
        self.reply("ok %s" % txt)

    def parse(self, txt=""):
        """ parse provided text or available self.txt and determine cmnd, args, rest and other values. Adds a _parsed object to the event. """
        if txt: self.txt = txt.rstrip()
        if self.txt.endswith("&"): self.threaded = True ; self.txt = self.txt[:-1]
        splitted = self.txt.split()
        self._parsed.args = []
        self._parsed.counter = -1
        self._parsed.fields = []
        self._parsed.words = []
        self._parsed.rest = ""
        quoted = False
        key2 = ""
        for word in splitted:
            self._parsed.counter += 1
            if self._parsed.counter == 0:
                if "cc" in self and self.cc == word[0]:
                    word = word[1:]
                    self._parsed.cmnd = word.lower()
                else: self._parsed.cmnd = word.lower()
                self._parsed.cmnd = self._parsed.cmnd.strip()
                continue
            try: key, value = word.split("=", 1) 
            except (IndexError,ValueError): key = "" ; value = word 
            if value.startswith('"'):
                if value.endswith('"'): value = value[1:-1] ; self._parsed.words.append(value)
                else: key2 = key ; value = value[1:] ; self._parsed.words.append(value) ; quoted = True ; continue
            if quoted:
                if value.endswith('"'): key = key2 ; self._parsed.words.append(value[:-1]) ; value = sj(*self._parsed.words) ; self._parsed.words = [] ; quoted = False
                else: self._parsed.words.append(value) ; continue
            if quoted: self._parsed.words.append(value) ; continue
            if self._parsed.counter == 1:
                if word.count("[") == 1:
                    value, index = word.split("[")
                    try: self._parsed.index = int(index[:-1])
                    except ValueError: self._parsed.index = 0
                if not self.prefix:
                    if key and os.path.isdir(j(cfg.main.workdir, key)): self.prefix = key
                    elif os.path.isdir(j(cfg.main.workdir, value)): self.prefix = value
            if "http" in value:
                self._parsed.args.append(value)
                self._parsed.rest += " " + value
                continue
            if key == "start":
                self._parsed.start = to_time(value)
                continue 
            if key == "end":
                self._parsed.end = to_time(value)
                continue 
            if value.startswith("+") or value.startswith("-"):
                try: self._parsed.time_diff = int(value[1:])
                except ValueError: self._parsed.time_diff = 0
            if value.endswith("++") or value.endswith("--"):
                try: self._parsed.karma = value[:-2]
                except ValueError: pass
            if key and value:
                pre = key[0]
                op = key[-1]
                post = value[-1]
                last = value[-1]
                if post == "-": value = value[:-1]
                if key.startswith("!"): key = key[1:] ; self._parsed.switch[key] = value ; continue
                if op == "-": key = key[:-1] ; self._parsed.notwant[key] = value
                else: self._parsed.want[key] = value
                if post == "-" : continue
                self._parsed.fields.append(key)
                self._parsed.args.append(key)
                self._parsed.rest += " " + key
            else:
                self._parsed.fields.append(value)
                self._parsed.args.append(value)
                self._parsed.rest += " " + str(value)
        self._parsed.rest = self._parsed.rest.strip()

    def reply(self, txt):
        """ give a reply to the origin of this event. """
        if type(txt) in [list, tuple]: txt = ",".join([str(x) for x in txt])
        if "outer" in self:
            self.outer.write(str(txt) + "\n")
            self.outer.flush()
            return
        self.say(self.channel, str(txt))

    def say(self, channel, txt):
        """ say something on a channel, using the bots available in the fleet. """
        from meds.core import fleet
        if self.doresult:
            self._result.append(txt)
            return
        for bot in fleet:
            if self.btype not in str(type(bot)): continue
            bot.say(channel, txt)
