# meds/kernel.py
#
#

""" central piece of code that loads the plugins and starts services. """


from meds.utils.cli import set_completer, enable_history, termsetup, termreset
from meds.utils.misc import include, locked
from meds.utils.trace import get_exception
from meds.utils.name import name, sname
from meds.utils.join import j

from meds.scheduler import Scheduler
from meds.engine import Engine
from meds.object import Object
from meds.event import Event
from meds.ool import OOL
from meds.cfg import cfg

import meds.core

import importlib
import logging
import pkgutil
import termios
import types
import time
import tty
import sys

class Kernel(Engine):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.booted = False

    def announce(self, txt):
        from meds.core import fleet
        for bot in fleet:
            if bot == self: continue
            bot.announce(txt)

    def boot(self):
        if self.booted: logging.warn("# kernel is already booted.") ; return
        from meds.core import fleet, aliases
        self._resume.fd = sys.stdin.fileno()
        self._resume.old = termsetup(sys.stdin.fileno())
        self._connected.ready()
        self.walk("meds.cbs", True)
        for name in cfg.main.packages: self.walk(name)
        try: aliases.load(j(cfg.main.workdir, 'aliases'))
        except FileNotFoundError: pass
        if cfg.main.init or cfg.main.all:
            for m in cfg.main.modules.split(","):
                if m: self.load(m)
            thrs = self.initialize()
            if thrs: self.waiter(thrs)
        set_completer(sorted(set([cmnd for cmnd in self.list()])))
        enable_history()
        self.booted = True
        self.ready()

    def dispatch(self, event):
        from meds.core import aliases, names
        event.parse()
        logging.info("! dispatch %s" % event.origin)
        if event._funcs:
            try: event.dispatch()
            except: logging.error(get_exception())
            event.ready()
            return
        if not event._parsed.cmnd: event.ready() ; return
        txt = aliases.get(event._parsed.cmnd, "")
        if txt:
            event.txt = event.txt.replace(event._parsed.cmnd, txt)
            event.parse()
        funcs = self._handlers.get(event._parsed.cmnd)
        if not funcs:
            modnames = names.get(event._parsed.cmnd, [])
            for modname in modnames: self.load(modname)
            funcs = self._handlers.get(event._parsed.cmnd)
        if funcs:
            for func in funcs:
                if "threaded" in dir(func) and func.threaded:
                    logging.info("! launch %s" % str(func))
                    self.launch(func, event)
                else:
                    logging.info("! exec %s" % str(func))
                    try: func(event)
                    except: logging.error(get_exception())
        event.ready()

    def find(self, cmnd):
        from meds.core import names
        for name in names.get(cmnd, []):
            if name not in self._table: continue
            mod = self._table[name]
            obj = getattr(mod, cmnd, None)
            if obj and type(obj) == types.FunctionType:
                if "event" in obj.__code__.co_varnames:
                    yield obj

    def initialize(self, modname=""):
        modnames = self.walk("meds.run")
        thrs = []
        event = Event()
        event.parse()
        for name in modnames:
            n = name.split(".")[-1]
            mod = self._table.get(name, None)
            if not mod: mod = self.load(name)
            if n in cfg.main.default: mod.init(event) ; event.ready()
            if n not in cfg.main.exclude.split(','):
                thr = self.launch(mod.init, event, name="%s.initialize" % n)
                thrs.append(thr)
        return thrs

    def list(self, want=""):
        for modname in self.modules():
            mod = self.direct(modname)
            for key in dir(mod):
                if key in ["init", "shutdown"]: continue
                if want and key != want: continue
                obj = getattr(mod, key, None)
                if obj and type(obj) == types.FunctionType:
                    if "event" in obj.__code__.co_varnames:
                        yield key

    def resume(self): pass

    def say(self, channel, txt): print(txt)

    @locked
    def shutdown(self, close=True):
        logging.warn("# shutdown")
        from meds.core import fleet
        for bot in fleet:
            try: bot.stop()
            except AttributeError: pass
            except: logging.error(get_exception())
        termreset(self._resume.fd, self._resume.old)
